class VersantError(Exception):
    pass


class BadRequestError(VersantError):
    pass


class WrongPasswordError(VersantError):
    pass


class WrongCredentialsError(VersantError):
    pass


class BrowserError(VersantError):
    pass



class GroupedRow:
    def __init__(
        self,
        customer: str,
        service: str,
        description: str,
        rate: float,
        processing_fee: bool,
        manager: str,
        row_index_start: int,
        service_type: str = "",
    ):
        self.customer: str = customer
        self.service: str = service
        self.description: str = description
        self.rate: float = rate
        self.processing_fee: bool = processing_fee
        self.manager: str = manager
        self.row_index_start: int = row_index_start
        self.service_type: str = service_type



import logging
import requests
import SeleniumLibrary.errors
from ..basic.core import SitesCore
from ta_captcha_solver.ta_captcha_solver import TACaptchaSolver
from ta_captcha_solver.exceptions import NoCaptchaException
from datetime import date, datetime, timedelta
from .exceptions import CentralReachException, ScheduledMaintenance
import time
import os
import shutil
import contextlib
import pyotp
from retry import retry


class CentralReachCore(SitesCore):
    def __init__(
        self,
        login: str,
        password: str,
        auth_totp_key: str = "",
        captcha_api_key: str | (str, str) = "",
        url: str = "https://members.centralreach.com/#login",
        start_date: date = date.today().replace(day=1),
        end_date: date = date.today(),
        timeout: int = 30,
        temp_folder: str = "",
        headless: bool = False,
        client_id: str = "",
        client_secret: str = "",
    ):
        """
        CentralReachCore object. Please Inheritance it.

        :param login: login for CentralReach site.
        :param password: password for CentralReach site.
        :param url: url for CentralReach site.
        :param otp_code: otp_code for CentralReach site.
        :param auth_totp_key: auth_totp_key for CentralReach site.
        :param captcha_api_key: api_key for captcha solver - can be just api_key itself or (provider, api_key) tuple.
        :param start_date: start_date for filters. The default value is first day of the month.
        :param end_date: end_date for filters. The default value is today.
        :param timeout: timeout for action. The default value is 30 seconds.
        :param temp_folder: Path to temp folder. The default value is temp/cr.
        """
        super().__init__(login, password, auth_totp_key, url, timeout, temp_folder, headless)
        self.start_date: date = start_date
        self.end_date: date = end_date
        self.client_id = client_id
        self.client_secret = client_secret
        self._do_auth = bool(self.client_id) and bool(self.client_secret)
        if isinstance(captcha_api_key, str):
            captcha_provider = "captcha.guru"
            api_key = captcha_api_key
        else:
            captcha_provider = captcha_api_key[0]
            api_key = captcha_api_key[1]

        self.captcha_solver = TACaptchaSolver.get(
            captcha_type="v2",
            browser=self.browser,
            service_provider_name=captcha_provider,
            service_provider_key=api_key,
        )

    def _get_auth_token(self, client_id, client_secret):
        """Retrieve the authentication token."""
        data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "scope": "cr-api",
        }

        return requests.post("https://login.centralreach.com/connect/token", data=data)

    def _get_internal_cookie(self, jwt):
        """Exchange the authentication token for an internal session cookie."""
        import requests

        headers = {
            "Accept": "application/json",
        }

        json_data = {
            "token": jwt["access_token"],
        }

        return requests.post(
            "https://members.centralreach.com/api/?framework.authtoken", headers=headers, json=json_data
        )

    def login_to_site(self) -> bool:
        """
        Try login to the site.

        :return: Status of the login. True if successfully log in
        """
        username_field = '//*[@id="Username"]'
        next_btn = '//*[@id="login"]'
        password_field = '//*[@id="Password"]'

        self.is_site_available = False
        self.browser.close_browser()

        if self.temp_folder:
            if os.path.exists(self.temp_folder):
                shutil.rmtree(self.temp_folder)
            os.mkdir(self.temp_folder)
            self.browser.set_download_directory(self.temp_folder, True)

        for i in range(1, 4):
            try:
                self.browser.open_chrome_browser(self.url, headless=self._headless)

                # get API auth cookies for RPA automation
                if self._do_auth:
                    new_cookies = self._get_internal_cookie(
                        self._get_auth_token(self.client_id, self.client_secret).json()
                    )

                    for cookie_name, cookie_value in new_cookies.cookies.get_dict().items():
                        self.browser.add_cookie(str(cookie_name), str(cookie_value))

                self.browser.set_window_size(1920, 1080)
                self.wait_element(username_field, is_need_screenshot=False, timeout=10)

                self.check_scheduled_maintenance()

                self.wait_element(username_field)
                self.browser.input_text_when_element_is_visible(username_field, self.login)
                self.browser.click_button_when_visible(next_btn)

                self.wait_element(password_field)
                self.browser.input_text_when_element_is_visible(password_field, self.password)

                for attempt in range(1, 4):
                    try:
                        with contextlib.suppress(NoCaptchaException):
                            self.captcha_solver.solve()
                        self.browser.click_button_when_visible(next_btn)
                        logging.info("Successfully Bypassed Captcha")
                        break
                    except Exception as ex:
                        logging.error(f"Failed to Bypass Captcha Attempt: {attempt}")
                        if attempt == 3:
                            raise ex
                        continue

                self.is_password_expired = self.check_password_expired()
                if self.is_password_expired:
                    break

                if self.browser.does_page_contain_element(
                    '//div[@class="loginError"]/div[text()="There was an unexpected problem signing in. If you keep seeing this, please contact CentralReach Support"]'  # noqa
                ):
                    elem = self.browser.find_element(
                        '//div[@class="loginError"]/div[text()="There was an unexpected problem signing in. If you keep seeing this, please contact CentralReach Support"]'  # noqa
                    )
                    if elem.is_displayed():
                        logging.warning("Logging into CentralReach failed. There was an unexpected problem signing in.")
                        raise CentralReachException(
                            "Unexpected error during signing in. If you keep seeing this, contact CentralReach Support"
                        )

                self.pass_mfa()
                self.wait_element_and_refresh('//span[text()="Favorites"]', is_need_screenshot=False)

                if self.does_element_displayed('//span[text()="Continue To Login"]'):
                    self.click('//span[text()="Continue To Login"]')

                self.is_site_available = self.browser.does_page_contain_element('//span[text()="Favorites"]')
                self.base_url: str = self.get_base_url(self.browser.get_location())

                return self.is_site_available
            except ScheduledMaintenance as ex:
                raise ex
            except Exception as ex:
                logging.error(f"Logging into CentralReach. Attempt #{i} failed")
                logging.error(str(ex))
                self.browser.capture_page_screenshot(os.path.join(self.output_folder, f"Login_failed_Attempt_{i}.png"))
                self.browser.close_browser()
        return False

    def wait_element(self, xpath: str, timeout: int = 60, is_need_screenshot: bool = True) -> bool:
        """
        Wait element some time.

        :param xpath: Xpath of the element
        :param timeout: How long to wait if the item does not appear?
        :param is_need_screenshot: Do need to take a screenshot?
        :return: True if element found, else False
        """
        is_success: bool = False
        timer: datetime = datetime.now() + timedelta(seconds=timeout)

        while not is_success and timer > datetime.now():
            self.__close_specific_windows('//button[contains(@id, "pendo-close-guide")]')
            self.__close_specific_windows('//button[text()="Okay, got it!"]')
            self.__close_specific_windows('//button[text()="Remind Me Later"]')
            self.__close_specific_windows('//button[text()="REGISTER NOW"]')

            self.check_scheduled_maintenance()

            if self.browser.does_page_contain_element(xpath):
                try:
                    is_success = self.browser.find_element(xpath).is_displayed()
                except Exception:
                    time.sleep(1)

            if not is_success:
                if self.browser.does_page_contain_element(
                    "//div[@id='select2-drop']/ul[@class='select2-results']/li[@class='select2-no-results']"
                ):
                    elem = self.browser.find_element(
                        "//div[@id='select2-drop']/ul[@class='select2-results']/li[@class='select2-no-results']"
                    )
                    if elem.is_displayed():
                        break
        if not is_success and is_need_screenshot:
            now: datetime = datetime.now()
            logging.warning(f'[{now.strftime("%H:%M:%S")}] Element \'{xpath}\' not available')
            self.browser.capture_page_screenshot(
                os.path.join(self.output_folder, f'Element_not_available_{now.strftime("%H_%M_%S")}.png')
            )
        return is_success

    def __close_specific_windows(self, xpath: str) -> None:
        if self.browser.does_page_contain_element(xpath):
            for i in range(5):
                try:
                    elements: list = self.browser.find_elements(xpath)
                    for element in elements:
                        try:
                            if element.is_displayed():
                                logging.warning(
                                    "A pop-up appeared and the bot closed it. "
                                    "Please validate the screenshot in the artifacts."
                                )
                                self.browser.capture_page_screenshot(
                                    os.path.join(
                                        self.output_folder, f'Pop_up_{datetime.now().strftime("%H_%M_%S")}.png'
                                    )
                                )
                                element.click()
                                self.browser.wait_until_element_is_not_visible(
                                    f"({xpath})[{elements.index(element) + 1}]"
                                )
                        except (AssertionError, SeleniumLibrary.errors.ElementNotFound):
                            time.sleep(1)
                    return None
                except Exception as ex:
                    logging.warning(ex)

    def check_scheduled_maintenance(self):
        if self.browser.does_page_contain_element("//div[contains(text(),'We’ll Be Back Soon')]"):
            if self.browser.find_element("//div[contains(text(),'We’ll Be Back Soon')]").is_displayed():
                self.browser.capture_page_screenshot(
                    os.path.join(self.output_folder, f'Maintenance_{datetime.now().strftime("%H_%M_%S")}.png')
                )
                raise ScheduledMaintenance(
                    "Data processing stopped due to Scheduled Maintenance. "
                    "Please run the bot again when the CentralReach is available"
                )

    def wait_element_and_refresh(self, xpath: str, timeout: int = 120, is_need_screenshot: bool = True) -> bool:
        """
        Wait element some time and refresh page.

        :param xpath: Xpath of the element
        :param timeout: How long to wait if the item does not appear?
        :param is_need_screenshot: Do need to take a screenshot?
        :return: True if element found, else False
        """
        is_success: bool = False
        timer: datetime = datetime.now() + timedelta(seconds=timeout)

        while not is_success and timer > datetime.now():
            is_success = self.wait_element(xpath, 45, False)
            if not is_success:
                self.browser.reload_page()
        if not is_success and is_need_screenshot:
            now: datetime = datetime.now()
            logging.warning(f'[{now.strftime("%H:%M:%S")}] Element \'{xpath}\' not available')
            self.browser.capture_page_screenshot(
                os.path.join(self.output_folder, f'Element_not_available_{now.strftime("%H_%M_%S")}.png')
            )
        return is_success

    def apply_filter(self, filter_name: str, additional_params: str = "", is_need_wait: bool = True):
        if self.browser.does_page_contain_element('//li/a[@data-click="openMenu"]'):
            if self.browser.find_element('//li/a[@data-click="openMenu"]').is_displayed():
                self.click('//li/a[@data-click="openMenu"]')

        self.wait_element("//li/a[text()='Filters']")
        self.browser.click_element_when_visible("//li/a[text()='Filters']")

        self.wait_element("//a/span[text()='Saved filters']")
        if self.browser.does_page_contain_element("//a/span[text()='" + filter_name + "']"):
            if not self.browser.find_element("//a/span[text()='" + filter_name + "']").is_displayed():
                self.browser.click_element_when_visible("//a/span[text()='Saved filters']")
        else:
            self.browser.click_element_when_visible("//a/span[text()='Saved filters']")
        self.wait_element("//a/span[text()='" + filter_name + "']")
        self.browser.click_element_when_visible("//a/span[text()='" + filter_name + "']")
        self.wait_element("//li[contains(@class, 'filter-highlight')]")

        # Update filters value
        if additional_params:
            updated_url: str = self.browser.get_location()
            updated_url += additional_params
            self.browser.go_to(updated_url)

        if is_need_wait:
            self.browser.wait_until_element_is_not_visible("//em[contains(., '<loading>')]", timedelta(seconds=45))
            self.browser.wait_until_element_is_not_visible(
                "//div[contains(@data-bind, 'loading()')]", timedelta(seconds=45)
            )

    def is_no_results(self) -> bool:
        self.wait_element(
            '//div[text()="No results matched your keywords, filters, or date range" and not(@style="display: none;")]',
            5,
            False,
        )
        if self.browser.does_page_contain_element(
            '//div[text()="No results matched your keywords, filters, or date range" and not(@style="display: none;")]'
        ):
            return True
        return False

    def apply_label(self, label_add: str, label_remove: str = "") -> None:
        self.wait_element('//button[contains(., "Label selected")]')
        self.browser.click_element_when_visible('//button[contains(., "Label selected")]')

        if label_add:
            self.wait_element('//h4[text()="Apply Labels"]/../../div/div/ul/li/input')
            self.browser.input_text('//h4[text()="Apply Labels"]/../../div/div/ul/li/input', label_add)
            self.wait_element(f'//div[text()="{label_add}" and @role="option"]')
            if self.browser.does_page_contain_element(
                "//div[@id='select2-drop']/ul[@class='select2-results']/li[@class='select2-no-results']"
            ):
                self.browser.input_text('//h4[text()="Apply Labels"]/../../div/div/ul/li/input', label_add.lower())
                self.wait_element(f'//div[text()="{label_add}" and @role="option"]')
            self.browser.click_element_when_visible(f'//div[text()="{label_add}" and @role="option"]')
        if label_remove:
            self.wait_element('//h4[text()="Remove Labels"]/../../div/div/ul/li/input')
            self.browser.input_text('//h4[text()="Remove Labels"]/../../div/div/ul/li/input', label_remove)
            self.wait_element(f'//div[text()="{label_remove}" and @role="option"]')
            if self.browser.does_page_contain_element(
                "//div[@id='select2-drop']/ul[@class='select2-results']/li[@class='select2-no-results']"
            ):
                self.browser.input_text('//h4[text()="Remove Labels"]/../../div/div/ul/li/input', label_remove.lower())
                self.wait_element(f'//div[text()="{label_remove}" and @role="option"]')
            self.browser.click_element_when_visible(f'//div[text()="{label_remove}" and @role="option"]')

        self.browser.click_element_when_visible('//button[text()="Apply Label Changes"]')
        self.browser.wait_until_element_is_not_visible('//button[text()="Apply Label Changes"]')
        self.browser.wait_until_element_is_not_visible('//h2[text()="Bulk Apply Labels"]')

    def change_billing(self, column_name: str, billing: str):
        self.browser.click_element_when_visible(
            '//th[contains(.,"'
            + column_name
            + '")]/div/a[contains(@data-bind, "{title: \'Search for different provider\'}")]'
        )

        self.wait_element('//th[contains(.,"{column_name}")]/div/div/input')
        self.browser.input_text_when_element_is_visible(f'//th[contains(.,"{column_name}")]/div/div/input', billing)

        self.wait_element(f'//li/div[contains(.,"{billing}")]', timeout=10)
        if self.browser.does_page_contain_element(f'//li/div[contains(.,"{billing}")]'):
            self.browser.click_element_when_visible(f'//li/div[contains(.,"{billing}")]')
        self.sync_to_all_claims(column_name)

    def select_timesheets(self, timesheets_id: list) -> None:
        for timesheet_id in timesheets_id:
            self.browser.scroll_element_into_view(f'//tr[@id="billing-grid-row-{timesheet_id}"]/td/input')
            self.browser.select_checkbox(f'//tr[@id="billing-grid-row-{timesheet_id}"]/td/input')

    def unselect_timesheets(self, timesheets_id: list) -> None:
        for timesheet_id in timesheets_id:
            self.browser.scroll_element_into_view(f'//tr[@id="billing-grid-row-{timesheet_id}"]/td/input')
            self.browser.unselect_checkbox(f'//tr[@id="billing-grid-row-{timesheet_id}"]/td/input')

    def clear_location(self):
        self.click_button_by_text("Facility", "Clear location")
        self.sync_to_all_claims("Facility")

    def clear_provider_supplier(self):
        self.click_button_by_text("Provider Supplier", "Clear provider supplier")
        self.sync_to_all_claims("Provider Supplier")

    def clear_referrer(self):
        self.click_button_by_text("Referrer", "Clear referrer")
        self.sync_to_all_claims("Referrer")

    def click_button_by_text(self, column_name: str, button_text: str):
        self.wait_element(f'//th[contains(.,"{column_name}")]/div/div/a/i')
        self.browser.click_element_when_visible(f'//th[contains(.,"{column_name}")]/div/div/a/i')
        self.wait_element(f'//th[contains(.,"{column_name}")]//a[text()="{button_text}"]')
        self.browser.click_element_when_visible(f'//th[contains(.,"{column_name}")]//a[text()="{button_text}"]')

    def sync_to_all_claims(self, column_name: str):
        self.wait_element(f'//th[contains(.,"{column_name}")]/div/div/a/i')
        self.browser.click_element_when_visible(f'//th[contains(.,"{column_name}")]/div/div/a/i')
        self.wait_element(f'//th[contains(.,"{column_name}")]//a[text()="To all claims"]')
        self.browser.click_element_when_visible(f'//th[contains(.,"{column_name}")]//a[text()="To all claims"]')

    def click_action_and_bulk_merge_claims(self) -> None:
        self.wait_element('//a[contains(., "Action")]')
        self.browser.click_element_when_visible('//a[contains(., "Action")]')

        self.wait_element('//a[contains(., "Bulk-merge Claims")]')
        self.browser.click_element_when_visible('//a[contains(., "Bulk-merge Claims")]')

    def check_and_update_billing(self, valid_billing: str, full_billing_name: str):
        all_billings = self.browser.find_elements('//a[contains(@data-bind, "billingName")]')
        for billing in all_billings:
            if valid_billing.lower() not in str(billing.text).lower():
                self.change_billing("Billing", full_billing_name)
                break

    def check_password_expired(self):
        with contextlib.suppress(AssertionError):
            self.browser.wait_until_element_is_visible("//*[contains(text(), 'Your password has expired')]")
            logging.info("Please reset AI worker password for Central Reach")
            return True
        return False

    def reset_password(self, new_password):
        self.wait_element('//input[@id="NewPassword"]')
        self.browser.input_text_when_element_is_visible('//input[@id="NewPassword"]', self.login)

        self.wait_element('//input[@id="ConfirmNewPassword"]')
        self.browser.input_text_when_element_is_visible('//input[@id="ConfirmNewPassword"]', self.login)

        self.wait_element('//button[@data-testid="update-password"]')
        self.browser.click_element_when_visible('//button[@data-testid="update-password"]')

        self.browser.close_all_browsers()

    @retry(tries=3, delay=10)
    def pass_mfa(self):
        if not self.browser.is_element_visible('//*[@id="verify-code"]'):
            return
        if str(self.auth_totp_key):
            try:
                base32_key = self.auth_totp_key.replace(" ", "").upper()
                totp = pyotp.TOTP(base32_key)
                otp_code = totp.now()
                self.wait_element('//*[@id="Code"]')
                self.browser.input_text('//*[@id="Code"]', otp_code, clear=True)
                self.browser.click_button_when_visible('//*[@id="verify-code"]')
                time.sleep(5)
                if self.browser.is_element_visible('//*[contains(text(),"Invalid code")]'):
                    raise CentralReachException("Invalid MFA code")
                self.browser.wait_until_element_is_not_visible('//*[@id="Code"]')
            except Exception as ex:
                raise CentralReachException(f"Error during MFA: {str(ex)}")
        else:
            raise CentralReachException("Auth Key is not provided")



import json
import time
from datetime import timedelta, datetime

import requests
from RPA.Browser.Selenium import Selenium
from requests import Response
from retry import retry

from .exceptions import BadRequestError, AlwaysCareOtpCodeError


def retry_if_bad_request(func):
    attempt = 1
    tries = 3

    @retry(exceptions=BadRequestError, tries=tries, delay=1, backoff=2)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BadRequestError as ex:
            nonlocal attempt
            print(f"Bad request Attempt {attempt}...", "WARN")
            attempt = attempt + 1 if attempt < tries else 1
            raise ex

    return wrapper


class AlwaysCareRequestsCore:
    def __init__(self, login: str, password: str, url: str = "https://unumvisionpwp.skygenusasystems.com/PWP/Landing"):
        """
        AlwaysCareRequestsCore object. Please Inheritance it.

        :param login: login for CentralReach site.
        :param password: password for CentralReach site.
        """
        self.session = requests.session()
        # Login data
        self.__login = login
        self.__password = password
        self.__url = url

        self._session_xsrf_token = ""  # Get after login

        self._login_to_always_care(self.__login, self.__password, self.__url)

    def get_otp_code(self) -> str:
        """
        Please, modify this function in your code
        :return:
        """
        raise NotImplementedError("The method for getting the otp must be implemented in the descendant class")

    @retry((BadRequestError, AlwaysCareOtpCodeError), tries=2)
    def _login_to_always_care(self, login: str, password: str, url: str):
        browser = Selenium()
        try:
            browser.set_selenium_timeout(value=timedelta(seconds=60))
            browser.open_available_browser(url=url, headless=True)

            browser.wait_until_element_is_visible('//input[@id="UserName"]', timeout=timedelta(seconds=60))

            browser.input_text_when_element_is_visible('//input[@id="UserName"]', login)
            browser.input_text_when_element_is_visible('//input[@id="Password"]', password)
            browser.click_element_when_visible('//button[@id="LoginBlockButton"]')

            # Waiting for otp code page
            browser.wait_until_element_is_visible('//*[@id="VerificationCodeTextBox"]', timeout=timedelta(seconds=60))

            code = self.get_otp_code()
            browser.input_text_when_element_is_visible('//*[@id="VerificationCodeTextBox"]', code)
            browser.click_element_when_visible('//button[@id="SubmitButton"]')

            try:
                browser.wait_until_element_is_visible(
                    '//form[@id="eligibility-search-form"]', timeout=timedelta(seconds=60)
                )
            except AssertionError as ex:
                if browser.is_element_visible('//*[text()="Invalid verification code."]'):
                    raise AlwaysCareOtpCodeError("Invalid verification code for Always Care.")
                raise ex

            # Check and get cookies
            timeout_time = datetime.now() + timedelta(seconds=30)
            while datetime.now() < timeout_time:
                try:
                    # Set cookies
                    for name, value in browser.get_cookies(as_dict=True).items():
                        self.session.cookies.set(name, value)

                    # Save csrf-token
                    try:
                        self._session_xsrf_token = self.session.cookies["XSRF-TOKEN"]
                    except KeyError:
                        raise BadRequestError(
                            "Session doesn't contain the required cookie 'XSRF-TOKEN' after logging into Always Care"
                        )

                    # Check authorization
                    self.get_locations(re_authorize=False)
                    break
                except BadRequestError:
                    browser.reload_page()
                    browser.wait_until_element_is_visible('//*[text()="Invalid verification code."]')
                    time.sleep(5)
                    continue
            else:
                raise BadRequestError("Session is unauthorized after login to Always Care")
        finally:
            browser.close_browser()

    def _logout(self):
        if self._session_xsrf_token:
            self.session.get("https://unumvisionpwp.skygenusasystems.com/Account/LogOff")
            self._session_xsrf_token = ""

    def __del__(self):
        try:
            self._logout()
        except Exception:
            pass

    def get_headers(self, is_json=True, put_token=True, add_headers: dict = None) -> dict:
        """
        Prepare header object for request.

        :param is_json (bool): True if content-type should be json, else False.
        :param put_token (bool): True if 'csrf-token' should be added to headers, else False.
        :param add_headers (dict): dictionary with key-values that should be added to headers.
        """
        headers = {}
        if is_json:
            headers["Content-Type"] = "application/json;charset=UTF-8"

        if put_token:
            headers["X-XSRF-TOKEN"] = self._session_xsrf_token

        if add_headers:
            for key, value in add_headers.items():
                headers[key] = value
        return headers

    @staticmethod
    def _is_json_response(response) -> bool:
        try:
            result = response.json()
            if not isinstance(result, dict):
                return False
            return True
        except json.decoder.JSONDecodeError:
            return False

    def check_response(
        self, response: Response, mandatory_json: bool = False, exc_message: str = "", re_authorize: bool = True
    ) -> None:
        """
        This method check response and raise exception 'BadRequestError'
        If status code is 401 (unauthorized) then it will try login again
        :param response: response from request
        :param mandatory_json: bool, if True - it will check is response contain json data
        :param exc_message: text message which will be raise if response wrong
        :param re_authorize: bool, if True then it will try login again if status code is 401
        """

        exc_message = exc_message + "\n" if exc_message else ""

        def raise_bad_request_error():
            if self._is_json_response(response):
                raise BadRequestError(
                    f"{exc_message}Status Code: {response.status_code}, "
                    f"Json content: {response.json()}, Headers: {response.headers}"
                )
            else:
                raise BadRequestError(
                    f"{exc_message}Status Code: {response.status_code}, " f"Headers: {response.headers}"
                )

        if re_authorize and response.status_code == 401:
            self._login_to_always_care(self.__login, self.__password, self.__url)
            raise BadRequestError(
                f"{exc_message}Status Code: {response.status_code} (Unauthorized request), "
                f"Json content: {response.json()}, Headers: {response.headers}"
            )

        if response.status_code != 200 or (mandatory_json and not self._is_json_response(response)):
            if re_authorize:
                self._login_to_always_care(self.__login, self.__password, self.__url)
            raise_bad_request_error()

        # Always Care case
        if response.status_code == 200 and self._is_json_response(response):
            if "statusCode" in response.json() and response.json()["statusCode"] == 500:
                if re_authorize:
                    self._login_to_always_care(self.__login, self.__password, self.__url)
                raise_bad_request_error()

    @retry_if_bad_request
    def get_locations(self, re_authorize=True):
        payload = {
            "GET_LOCATION": True,
            "GET_PROVIDER": True,
            "USE_COOKIE_FOR_PROVIDER": True,
        }
        url = "https://unumvisionpwp.skygenusasystems.com/DropList/GetSearchCriteriaDropdowns"

        response = self.session.get(url=url, headers=self.get_headers(put_token=True), json=payload)

        exception_message = "Problems with getting locations."
        self.check_response(response, mandatory_json=True, exc_message=exception_message, re_authorize=re_authorize)

        return response.json()

    @retry_if_bad_request
    def get_member_eligibility_search(
        self,
        provider_id: int,
        location_id: int,
        birth_date: datetime,
        service_date: datetime,
        first_name: str = None,
        last_name: str = None,
        subscriber_id: str = None,
    ):
        payload = {
            "LAST_NAME": last_name,
            "FIRST_NAME": first_name,
            "DOB": birth_date.strftime("%m/%d/%Y"),
            "SUBSCRIBER_ID": subscriber_id,
            "DATE_OF_SERVICE": service_date.strftime("%m/%d/%Y"),
            "LOCATION_ID": location_id,
            "PROVIDER_ID": provider_id,
            "SSN": "",
            "Log": True,
        }
        url = "https://unumvisionpwp.skygenusasystems.com/Welcome/MemberEligibilitySearch/"
        response = self.session.post(url, headers=self.get_headers(put_token=True), json=payload)

        exception_message = "Problems with getting member eligibility search result."
        self.check_response(response, mandatory_json=True, exc_message=exception_message)

        return response.json()



import json
import time
from datetime import timedelta
from threading import Event, Thread

import requests
from RPA.Browser.Playwright import Playwright
from retry import retry
from fake_useragent import UserAgent

from .exceptions import VersantError, BadRequestError, WrongCredentialsError, BrowserError


def retry_if_bad_request(func):
    attempt = 1
    tries = 3

    @retry(exceptions=BadRequestError, tries=tries, delay=1, backoff=2)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BadRequestError as ex:
            nonlocal attempt
            print(f"Bad request Attempt {attempt}...", "WARN")
            attempt = attempt + 1 if attempt < tries else 1
            raise ex

    return wrapper


class VersantRequestsCore:
    def __init__(self, login: str, password: str, url: str = "https://ecp.versanthealth.com/prelogin/login"):
        """
        VersantSuperiorRequestsCore object. Please Inheritance it.

        :param login: login for CentralReach site.
        :param password: password for CentralReach site.
        """
        self.session = requests.session()
        # usable user-agent
        self.user_agent = UserAgent().chrome
        # Login data
        self.__login = login
        self.__password = password
        self.__url = url

        self._transaction_id = ""  # Get from login response
        self._user_id = login

        # Get from local storage
        self._tax_id = ""
        self._bearer_token = ""
        self._bearer_token_type = ""
        self._bearer_token_expired_in = ""  # TODO implement refresh
        self._bearer_refresh_token = ""  # TODO implement refresh

        self._login_to_versant(self.__login, self.__password, self.__url)

    @staticmethod
    def __setup_browser(open_as_headless=True):
        """
        Open new browser and apply settings
        """
        from Browser import SupportedBrowsers

        browser = Playwright()
        browser.set_browser_timeout(timedelta(seconds=100))
        browser.new_browser(SupportedBrowsers.chromium, headless=open_as_headless, timeout=timedelta(seconds=120))
        browser.new_context(userAgent=UserAgent().chrome)
        browser.new_page()
        return browser

    @staticmethod
    def _is_json_response(response) -> bool:
        try:
            response.json()
            return True
        except json.decoder.JSONDecodeError:
            return False

    def check_response(
        self,
        response,
        mandatory_json: bool = False,
        exc_message: str = "",
        re_authorize: bool = True,
        check_resp_obj: bool = False,
    ) -> None:
        """
        This method check response and raise exception 'BadRequestError'
        If status code is 401 (unauthorized) then it will try login again
        :param response: response from request
        :param mandatory_json: bool, if True - it will check is response contain json data
        :param exc_message: text message which will be raise if response wrong
        :param re_authorize: bool, if True then it will try login again if status code is 401
        """
        if re_authorize and response.status_code == 401:
            self._login_to_versant(self.__login, self.__password, self.__url)
            raise BadRequestError(
                f"{exc_message}Status Code: {response.status_code} (Unauthorized request), "
                f"Json content: {response.json()}, Headers: {response.headers}"
            )

        if response.status_code != 200 or (mandatory_json and not self._is_json_response(response)):
            exc_message = exc_message + "\n" if exc_message else ""
            if self._is_json_response(response):
                raise BadRequestError(
                    f"{exc_message}Status Code: {response.status_code}, "
                    f"Json content: {response.json()}, Headers: {response.headers}"
                )
            else:
                raise BadRequestError(
                    f"{exc_message}Status Code: {response.status_code}, " f"Headers: {response.headers}"
                )
        if check_resp_obj:
            response_json = response.json()
            if "body" not in response_json:
                raise VersantError("Response doesn't contain body")

            if "responseObject" not in response_json["body"]:
                if "messages" in response_json["body"] and response_json["body"]["messages"]:
                    messages = []
                    for message in response_json["body"]["messages"]:
                        messages.append(
                            f"{message['errorIdentifier'] if 'errorIdentifier' in message else '-'}: "
                            f"{message['description']}"
                        )
                    messages_str = "\n".join(messages)
                    raise VersantError(
                        "Response doesn't contain 'responseObject', " f"error messages: \n{messages_str}"
                    )
                else:
                    raise VersantError("Response doesn't contain 'responseObject'")

    @retry(BrowserError, tries=2)
    def _login_to_versant(self, login: str, password: str, url: str):
        browser = None
        try:
            try:
                from Browser.utils.data_types import PageLoadStates

                browser = self.__setup_browser()
                browser.go_to(url=url, timeout=timedelta(seconds=180), wait_until=PageLoadStates.domcontentloaded)

                browser.wait_for_elements_state('//*[@id="username"]', timeout=timedelta(seconds=120))
                browser.fill_text('//*[@id="username"]', login)
                browser.fill_text('//*[@id="password"]', password)

                # To catch the response of the login, we start to wait until the click of the Login button
                response = {}

                def wait_for_response(ev: Event):
                    nonlocal response
                    login_url = "https://api.versanthealth.com/AccessManagementSecured/api/AccessManagement/Login"
                    response = browser.wait_for_response(login_url, timeout=timedelta(seconds=60))
                    ev.set()

                event = Event()

                t1 = Thread(target=wait_for_response, args=(event,))  # Start wait for response
                t1.start()

                time.sleep(1)
                browser.click('//button[@class="btn btn-light submit-button"]')  # Click of the Login button
                is_event_succ = event.wait(600.0)  # Wait while response is ready

                if not is_event_succ:
                    raise BrowserError("Timeout of 600 seconds reached on login.")

            except Exception as ex:
                raise BrowserError(f"Can't login to Versant. Exception: {ex}") from ex

            response_body = response["body"]["body"]
            if response_body["statuscode"] != "OK":
                if response_body["messages"]:
                    error_message: dict = response_body["messages"][0]
                    if error_message["code"] == "11.01.05":
                        raise WrongCredentialsError(
                            "Versant login attempt failed due to incorrect credentials. "
                            f"Details: {error_message['description']}"
                        )
                raise VersantError("Can't login to Versant")

            self._transaction_id = response["body"]["header"]["transactionid"]

            self._tax_id = json.loads(browser.local_storage_get_item("userDetails"))["taxId"]
            self._bearer_token = json.loads(browser.local_storage_get_item("bearerToken"))["access_token"]
            self._bearer_token_type = json.loads(browser.local_storage_get_item("bearerToken"))["token_type"]
            self._bearer_token_expired_in = json.loads(browser.local_storage_get_item("bearerToken"))["expires_in"]
            self._bearer_refresh_token = json.loads(browser.local_storage_get_item("bearerToken"))["refresh_token"]

            if not self._bearer_token:
                raise BadRequestError(
                    "Session doesn't contain the required cookie 'bearer-token' after logging into Versant"
                )

            # Check authorization
            try:
                self.get_office_locations_by_tax_id()
            except BadRequestError:
                raise BadRequestError("Session is unauthorized after login to Versant")
        finally:
            if browser is not None:
                browser.close_browser("CURRENT")

    @retry_if_bad_request
    def get_office_locations_by_tax_id(self):
        url = "https://api.versanthealth.com/ProviderService/GetOfficeLocationsByTaxId"
        response = self.session.post(url, headers=self.get_headers(), json=self.get_payload())

        exception_message = "Problems with getting office by tax id."
        self.check_response(response, mandatory_json=True, exc_message=exception_message)
        return response.json()

    def get_payload(self, body_payload: dict = {}) -> dict:
        return {
            "header": {"transactionId": self._transaction_id, "userId": self._user_id, "taxId": self._tax_id},
            "body": body_payload,
        }

    def get_headers(self, is_json=True, put_token=True, add_headers: dict = None) -> dict:
        """
        Prepare header object for request.

        :param is_json (bool): True if content-type should be json, else False.
        :param put_token (bool): True if 'csrf-token' should be added to headers, else False.
        :param add_headers (dict): dictionary with key-values that should be added to headers.
        """
        headers = {}
        if is_json:
            headers["Content-Type"] = "application/json"

        if put_token:
            headers["Authorization"] = f"{self._bearer_token_type} {self._bearer_token}"

        if add_headers:
            for key, value in add_headers.items():
                headers[key] = value
        return headers



import logging
from ..basic.core import SitesCore
import os
import traceback
from datetime import timedelta


class SalesforceCore(SitesCore):
    """
    Salesforce object. Please Inheritance it.
    """

    def get_mfa_code(self) -> str:
        """
        Please, modify this function in your code
        :return:
        """

        return ""

    def get_email_code(self) -> str:
        """
        Please, modify this function in your code
        :return:
        """

        return ""

    def login_to_site(self) -> bool:
        self.is_site_available = False
        self.browser.close_browser()

        if self.temp_folder:
            if not os.path.exists(self.temp_folder):
                os.mkdir(self.temp_folder)
            self.browser.set_download_directory(self.temp_folder, True)

        for i in range(1, 4):
            try:
                self.browser.open_chrome_browser(self.url)
                self.browser.set_browser_implicit_wait(timedelta(seconds=30))
                self.browser.set_window_size(1920, 1080)
                # self.browser.maximize_browser_window()

                self.wait_element('//input[@id="username"]')
                self.browser.input_text('//input[@id="username"]', self.login)
                self.browser.input_text('//input[@id="password"]', self.password)
                self.browser.click_element('//input[@id="Login" and @type="submit"]')

                self.wait_element('//input[@id="tc"]', timeout=5, is_need_screenshot=False)
                if self.does_element_displayed('//input[@id="tc"]'):
                    self.browser.input_text('//input[@id="tc"]', self.get_mfa_code())
                elif self.does_element_displayed('//input[@id="emc"]'):
                    self.browser.input_text('//input[@id="emc"]', self.get_email_code())

                if self.does_element_displayed('//input[@id="save"]'):
                    self.browser.click_element('//input[@id="save"]')

                self.wait_element('//button[@aria-label="Search"]')
                if self.browser.does_page_contain_element('//button[@aria-label="Search"]'):
                    self.base_url = self.get_base_url(self.browser.get_location())
                    self.is_site_available = True
                    return True
            except Exception as ex:
                logging.error(f"Login failed. Attempt {i}")
                logging.error(str(ex))
                traceback.print_exc()
                self.browser.capture_page_screenshot(os.path.join(self.output_folder, f"Login_failed_Attempt_{i}.png"))
                self.browser.close_browser()
        return False



from .central_reach.core import CentralReachCore
from .central_reach.requests_core import CentralReachRequestsCore
from .central_reach.exceptions import (
    CentralReachException,
    ScheduledMaintenance,
    BadRequest,
    EmptyPage,
    CentralReachError,
    ScheduledMaintenanceError,
    EmptyPageError,
    BadRequestError,
)

__all__ = [
    "CentralReachCore",
    "CentralReachError",
    "ScheduledMaintenanceError",
    "EmptyPageError",
    "BadRequestError",
    "CentralReachRequestsCore",
    "CentralReachException",
    "ScheduledMaintenance",
    "BadRequest",
    "EmptyPage",
]



import datetime
import json
import time
import requests
from retry import retry
from .exceptions import BadRequestError, ScheduledMaintenanceError, PasswordExpiredException
from .core import CentralReachCore


def retry_if_bad_request(func):
    attempt = 1
    tries = 3

    @retry(exceptions=BadRequestError, tries=tries, delay=1, backoff=2)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BadRequestError as ex:
            nonlocal attempt
            print(f"Bad request Attempt {attempt}...", "WARN")
            attempt = attempt + 1 if attempt < tries else 1
            raise ex

    return wrapper


class CentralReachRequestsCore:
    def __init__(
        self,
        login: str,
        password: str,
        auth_totp_key: str = "",
        captcha_api_key: str | (str, str) = "",
        bw_creds: dict = None,
        client_id: str = "",
        client_secret: str = "",
    ):
        """
        CentralReachRequestsCore object. Please Inheritance it.

        :param login: login for CentralReach site.
        :param password: password for CentralReach site.
        :param captcha_api_key: api_key for captcha solver - can be just api_key itself or (provider, api_key) tuple.
        """
        self.__login = login
        self.__password = password
        self.__auth_totp_key = auth_totp_key
        self.__captcha_api_key = captcha_api_key
        self.__client_id = client_id
        self.__client_secret = client_secret
        self._session_csrf_token = ""
        self.session = requests.session()
        self._login_to_central_reach()

    def _get_headers(self, is_json=True, add_headers: dict = None, put_token: bool = True) -> dict:
        """
        Prepare header object for request.

        :param is_json (bool): True if content-type should be json, else False.
        :param add_headers (dict): dictionary with key-values that should be added to headers.
        :param put_token (bool): True if 'csrf-token' should be added to headers, else False.
        """
        headers = {"origin": "https://members.centralreach.com"}
        if is_json:
            headers["content-type"] = "application/json; charset=UTF-8"

        if put_token:
            headers["x-csrf-token"] = self._session_csrf_token

        if add_headers:
            for key, value in add_headers.items():
                headers[key] = value
        return headers

    @staticmethod
    def __is_scheduled_maintenance(response) -> bool:
        if response.status_code == 200 and "scheduled maintenance" in response.text.lower():
            return True
        return False

    @staticmethod
    def _is_json_response(response) -> bool:
        try:
            response.json()
            return True
        except json.decoder.JSONDecodeError:
            return False

    def check_response(
        self, response, mandatory_json: bool = False, exc_message: str = "", re_authorize: bool = True
    ) -> None:
        """
        This method check response and raise exception 'BadRequestError'
           or 'ScheduledMaintenanceError' with exc_message,
        If status code is 401 (unauthorized) then it will try login again
        :param response: response from request
        :param mandatory_json: bool, if True - it will check is response contain json data
        :param exc_message: text message which will be raise if response wrong
        :param re_authorize: bool, if True then it will try login again if status code is 401
        """
        # Check if site is under scheduled maintenance
        if self.__is_scheduled_maintenance(response):
            print(exc_message, "Error")
            print("'Central Reach' site is currently unavailable due to scheduled maintenance", "Error")
            raise ScheduledMaintenanceError

        # Check if response is unauthorized
        if response.status_code == 401:
            unauthorized_request = True
        elif self._is_json_response(response) and response.json().get("message", "") == "Not logged in.":
            unauthorized_request = True
        else:
            unauthorized_request = False

        if unauthorized_request:
            if re_authorize:
                self._login_to_central_reach()
            raise BadRequestError(
                f"{exc_message}Status Code: {response.status_code} (Unauthorized request), "
                f"Json content: {response.json()}, Headers: {response.headers}"
            )

        # Check if response is not 200 or not json
        if response.status_code != 200 or (mandatory_json and not self._is_json_response(response)):
            exc_message = exc_message + "\n" if exc_message else ""
            if self._is_json_response(response):
                raise BadRequestError(
                    f"{exc_message}Status Code: {response.status_code}, "
                    f"Json content: {response.json()}, Headers: {response.headers}"
                )
            else:
                raise BadRequestError(
                    f"{exc_message}Status Code: {response.status_code}, " f"Headers: {response.headers}"
                )

    @retry_if_bad_request
    def _login_to_central_reach(self) -> None:
        self.central_reach_core = CentralReachCore(
            self.__login,
            self.__password,
            self.__auth_totp_key,
            self.__captcha_api_key,
            headless=True,
            client_id=self.__client_id,
            client_secret=self.__client_secret,
        )
        self.central_reach_core.login_to_site()
        self.session = requests.session()

        password_expired = self.central_reach_core.is_password_expired
        if password_expired:
            raise PasswordExpiredException("Please reset AI worker password for Central Reach")

        # Check authorisation
        timeout_time = datetime.datetime.now() + datetime.timedelta(seconds=30)
        while datetime.datetime.now() < timeout_time:
            try:
                # Set cookies
                for name, value in self.central_reach_core.browser.get_cookies(as_dict=True).items():
                    self.session.cookies.set(name, value)

                # Save csrf-token
                try:
                    self._session_csrf_token = self.session.cookies["csrf-token"]
                except KeyError:
                    raise BadRequestError(
                        "Session doesn't contain the required cookie 'csrf-token' after logging " "into Central Reach"
                    )

                # Check authorization
                self.__insurance_eligibility_status()
                break
            except BadRequestError:
                time.sleep(1)
                continue
        else:
            self.central_reach_core.browser.close_browser()
            raise BadRequestError("Session is unauthorized after login to Central Reach")
        self.central_reach_core.browser.close_browser()

    def __insurance_eligibility_status(self) -> None:
        header = {"referer": "https://members.centralreach.com/"}
        url = "https://members.centralreach.com/crxapi/system/dictionary/InsuranceEligibilityStatus"
        response = self.session.get(url, headers=self._get_headers(is_json=True, add_headers=header))

        exception_message = "Problems with insurance eligibility request."
        self.check_response(response, mandatory_json=True, exc_message=exception_message, re_authorize=False)

    @retry_if_bad_request
    def _get_filters(self):
        payload = {
            "applicationSection": "billingmanager.billing",
        }
        url = "https://members.centralreach.com/api/?shared.loadFilters"
        response = self.session.post(url, json=payload, headers=self._get_headers(is_json=True))

        exception_message = "Problems with getting billings."
        self.check_response(response, mandatory_json=True, exc_message=exception_message)
        return response.json()["filters"]

    def get_filter_by_name(self, filter_name):
        filters = self._get_filters()
        for filter_data in filters:
            if str(filter_data["Name"]).strip() == filter_name:
                return json.loads(filter_data["filters"])
        else:
            raise Exception("Filter '{filter_name}' doesn't exist")

    @retry_if_bad_request
    def get_era_list(self, start_date: datetime = None, end_date: datetime = None):
        _start_date = start_date.strftime("%Y-%m-%d") if start_date else ""
        _end_date = end_date.strftime("%Y-%m-%d") if start_date else ""

        load_era_list_url = "https://members.centralreach.com/api/?claims.loadERAList"
        data = {
            "startDate": _start_date,
            "endDate": _end_date,
            "page": "1",
            "claimLabelId": "",
            "pageSize": "2000",
        }
        response = self.session.get(load_era_list_url, json=data, headers=self._get_headers(is_json=False))
        if response.status_code != 200:
            response = self.session.get(load_era_list_url, json=data, headers=self._get_headers(is_json=False))

        if "application/json" in response.headers.get("content-type"):
            if response.status_code == 200 and response.json().get("success", False) is True:
                return response.json()
            elif "message" in response.json():
                raise Exception(
                    f"Problems with getting 'Era List' from 'Central Reach' site. {response.json()['message']}"
                )
        raise Exception("Problems with getting 'Era List' from 'Central Reach' site.")

    def get_zero_pay_filter(self, start_date: datetime = None, end_date: datetime = None) -> dict:
        response = self.get_era_list(start_date, end_date)
        era_list_data = response["items"]

        # Zero Pay filter
        zero_pay_data: dict = {}
        for item in era_list_data:
            if item["PaymentAmount"] == 0.0:
                zero_pay_data[str(item["Id"])] = item
        return zero_pay_data

    def get_pr_filter(self, start_date: datetime = None, end_date: datetime = None) -> dict:
        response = self.get_era_list(start_date, end_date)
        era_list_data = response["items"]

        # PR filter
        pr_data: dict = {}
        for item in era_list_data:
            if (
                item["PaymentAmount"] == 0.0
                and item["PrAdjustmentTotal"] > 0
                and item["PiAdjustmentTotal"] == 0.0
                and item["Reconciled"] == "None"
            ):
                pr_data[str(item["Id"])] = item
        return pr_data

    def get_denial_filter(self, start_date: datetime = None, end_date: datetime = None) -> dict:
        response = self.get_era_list(start_date, end_date)
        era_list_data = response["items"]

        # Denial filter
        denial_data: dict = {}
        for item in era_list_data:
            if item["PaymentAmount"] == 0.0 and item["Reconciled"] == "None":
                denial_data[str(item["Id"])] = item
        return denial_data
